@@ -4,8 +4,8 @@ use Exporter ();
 use Apache::Constants qw(OK DECLINED);
 use Apache::SIG ();
 
-@Apache::EXPORT_OK = qw(exit warn fork forkoption);
-$Apache::VERSION = "1.22";
+@Apache::EXPORT_OK = qw(exit warn);
+$Apache::VERSION = "1.24";
 
 *import = \&Exporter::import;
 
@@ -19,13 +19,13 @@ if (caller eq "CGI::Apache") {
     bootstrap Apache $Apache::VERSION;
 }
 else {
-    eval { bootstrap Apache $Apache::VERSION };
+    if(exists $ENV{MOD_PERL}) {
+      require Apache::Server unless $mod_perl::UNIMPORT{'server'};
+      require Apache::Connection unless $mod_perl::UNIMPORT{'connection'};
+	bootstrap Apache $Apache::VERSION;
+    }
     Apache::SIG->set;
 }
-if($@) {
-    die "$@\n" if exists $ENV{MOD_PERL};
-    warn "warning: can't `bootstrap Apache $Apache::VERSION' outside of httpd\n";
-}
 
 if($ENV{MOD_PERL} && perl_hook("Sections")) {
     *Apache::ReadConfig:: = \%ApacheReadConfig::;
@@ -50,7 +50,7 @@ sub parse_args {
 
 sub content {
     my($r) = @_;
-    my $ct = $r->header_in("Content-type");
+    my $ct = $r->header_in("Content-type") || "";
     return unless $ct eq "application/x-www-form-urlencoded";
     my $buff;
     $r->read($buff, $r->header_in("Content-length"));
@@ -63,13 +63,8 @@ sub args {
 	       $val ? $r->query_string($val) : $r->query_string);
 }
 
-sub cgi_var {
-    my($r, $key) = @_;
-    my $val = $r->cgi_env($key);
-    return $val;
-}
+*READ = \&read unless defined &READ;
 
-*READ = \&read;
 sub read {
     my($r, $bufsiz) = @_[0,2];
     my($nrd, $buf, $total);
@@ -99,7 +94,53 @@ sub read {
     return $total;
 }
 
-sub GETC { my $c; shift->read($c,1); $c; }
+sub new_read {
+    my($r, $bufsiz) = @_[0,2];
+    my($nrd, $buf, $total);
+    $nrd = $total = 0;
+    $buf = "";
+    $_[1] ||= "";
+
+    if(my $rv = $r->setup_client_block) {
+	$r->log_error("Apache->read: setup_client_block returned $rv");
+	die $rv;
+    }
+
+    #XXX: must set r->read_length to 0 here,
+    #since this read() method may be called in loop
+    #in which case, the second time in, should_client_block() 
+    #thinks we've already read the request body and returns 0
+    $r->read_length(0); 
+
+    unless($r->should_client_block) {
+	my $rl = $r->read_length;
+	$r->log_error("Apache->read: should_client_block returned 0 (rl=$rl)");
+	return 0;
+    }
+
+    $r->hard_timeout("Apache->read");
+    
+    while($bufsiz) {
+	$nrd = $r->get_client_block($buf, $bufsiz) || 0;
+	if(defined $nrd and $nrd > 0) {
+	    $bufsiz -= $nrd;
+	    $_[1] .= $buf;
+ 	    #substr($_[1], $total, $nrd) = $buf;
+	    $total += $nrd;
+	    $r->reset_timeout;
+	    next if $bufsiz;
+	    last;
+	}
+	else {
+	    $_[1] = undef;
+	    last;
+	}
+    }
+    $r->kill_timeout;
+    return $total;
+}
+
+sub GETC { my $c; shift->READ($c,1); $c; }
 
 #shouldn't use <STDIN> anyhow, but we'll be nice
 sub READLINE { 
@@ -109,13 +150,18 @@ sub READLINE {
     $line;
 }
 
-*PRINT = \&print;
-
 sub PRINTF {
     my $r = shift;
     my $fmt = shift;
     $r->print(sprintf($fmt, @_));
 }
+*printf = \&PRINTF;
+
+sub WRITE {
+    my($r, $buff, $length, $offset) = @_;
+    my $send = substr($buff, $offset, $length);
+    $r->print($send);
+}
 
 sub send_cgi_header {
     my($r, $headers) = @_;
@@ -145,29 +191,6 @@ sub send_cgi_header {
     }
 }
 
-sub as_string {
-    my($r) = @_;
-    my($k,$v,@retval);
-    my(%headers_in) = $r->headers_in;
-
-    push @retval, $r->the_request;
-    while(($k,$v) = each %headers_in) {
-	push @retval, "$k: $v";
-    }
-
-    push @retval, "";
-
-    push @retval, $r->status_line;
-    for (qw(err_headers_out headers_out)) {
-	my(%headers_out) = $r->$_();
-
-	while(($k,$v) = each %headers_out) {
-	    push @retval, "$k: $v";
-	}
-    }    
-    join "\n", @retval, "";
-}
-
 sub TIEHANDLE {
     my($class, $r) = @_;
     $r ||= Apache->request;
@@ -237,6 +260,11 @@ structure or C<undef> if there is no previous request.
 This method returns a blessed reference to the next (internal) request
 structure or C<undef> if there is no next request.
 
+=item $r->last
+
+This method returns a blessed reference to the last (internal) request
+structure.  Handy for logging modules.
+
 =item $r->is_main
 
 Returns true if the current request object is for the main request.
@@ -384,15 +412,9 @@ returned.  When called in a list context, a list of parsed I<key> =>
 I<value> pairs are returned.  *NOTE*: you can only ask for this once,
 as the entire body is read from the client.
 
-=item $r->read_client_block($buf, $bytes_to_read)
-
-Read from the entity body sent by the client.  Example of use:
-
-   $r->read_client_block($buf, $r->header_in('Content-length'));
-
 =item $r->read($buf, $bytes_to_read)
 
-This method uses read_client_block() to read data from the client, 
+This method is used to read data from the client, 
 looping until it gets all of C<$bytes_to_read> or a timeout happens.
 
 In addition, this method sets a timeout before reading with
@@ -449,6 +471,10 @@ The dotted decimal representation of the remote client's IP address.
 This is set by then server when the connection record is created so
 is always defined.
 
+You can also set this value by providing an argument to it. This is
+helpful if your server is behind a squid accelerator proxy which add
+a HTTP_X_FORWARDED_FOR header.
+
 =item $c->local_addr
 
 A packed SOCKADDR_IN in the same format as returned by
@@ -577,6 +603,10 @@ module provides the constants to check against.
 		      $filename);
    }
 
+=item $r->get_server_port
+
+Returns the port number on which the server is listening.
+
 =item $s = $r->server
 
 Return a reference to the server info object (blessed into the
@@ -606,7 +636,7 @@ Returns true if this is a virtual server.
 
 =item $s->names
 
-Returns the wild-carded names for HostAlias servers. 
+Returns the wild-carded names for ServerAlias servers. 
 
 =item $s->warn
 
@@ -706,10 +736,11 @@ Get or set the content encoding.  Content encodings are string like
 "gzip" or "compress".  This correspond to the "Content-Encoding"
 header in the HTTP protocol.
 
-=item $r->content_language( [$newval] )
+=item $r->content_languages( [$array_ref] )
 
-Get or set the content language.  The content language corresponds to the
-"Content-Language" HTTP header and is a string like "en" or "no".
+Get or set the content languages.  The content language corresponds to the
+"Content-Language" HTTP header and is an array reference containing strings
+such as "en" or "no".
 
 =item $r->status( $integer )
 
@@ -777,12 +808,16 @@ like this:
   $r->send_fd(FILE);
   close(FILE);
 
-=item $r->internal_redirect_handler( $newplace )
+=item $r->internal_redirect( $newplace )
 
 Redirect to a location in the server namespace without 
 telling the client. For instance:
 
-   $r->internal_redirect_handler("/home/sweet/home.html");
+   $r->internal_redirect("/home/sweet/home.html");
+
+=item $r->internal_redirect_handler( $newplace )
+
+Same as I<internal_redirect>, but the I<handler> from C<$r> is preserved.
 
 =item $r->custom_response($code, $uri)
 
@@ -864,6 +899,8 @@ kill_timeout() will disarm either variety of timeout.
 
 reset_timeout() resets the timeout in progress.
 
+=item $r->post_connection($code_ref)
+
 =item $r->register_cleanup($code_ref)
 
 Register a cleanup function which is called just before $r->pool is
@@ -874,6 +911,10 @@ destroyed.
        warn "registered cleanup called for ", $r->uri, "\n";
    });
 
+The I<post_connection> method is simply an alias for I<register_cleanup>, 
+as this method may be used to run code after the client connection is closed,
+which may not be a I<cleanup>.
+
 =back
 
 =head1 CGI SUPPORT
@@ -883,34 +924,6 @@ type of interface.
 
 =over 4
 
-=item $r->cgi_env
-
-Return a %hash that can be used to set up a standard CGI environment.
-Typical usage would be:
-
-   %ENV = $r->cgi_env
-
-B<NOTE:> The $ENV{GATEWAY_INTERFACE} is set to C<'CGI-Perl/1.1'> so
-you can say:
-
-   if($ENV{GATEWAY_INTERFACE} =~ /^CGI-Perl/) {
-       # do mod_perl stuff
-   }
-   else {
-       # do normal CGI stuff
-   }
-
-When given a key => value pair, this will set an environment variable.
-
-   $r->cgi_env(REMOTE_GROUP => "camels");
-
-=item $r->cgi_var($key);
-
-Calls $r->cgi_env($key) in a scalar context to prevent the mistake
-of calling in a list context.
-
-   my $doc_root = $r->cgi_env('DOCUMENT_ROOT');
-
 =item $r->send_cgi_header()
 
 Take action on certain headers including I<Status:>, I<Location:> and
@@ -992,7 +1005,7 @@ Apache C API notes at C<http://www.apache.org/docs/>
 
 Perl interface to the Apache C API written by Doug MacEachern
 with contributions from Gisle Aas, Andreas Koenig, Eric Bartley, 
-Rob Hartill, Gerald Richter, Salvador Garcia and others. 
+Rob Hartill, Gerald Richter, Salvador Ortiz and others. 
 
 =cut
 
