@@ -1,85 +1,20 @@
 package Apache::Constants;
 
-$Apache::Constants::VERSION = "1.08";
-
-use Exporter ();
-use strict;
-
-*import = \&Exporter::import;
+$Apache::Constants::VERSION = "1.09";
 
 unless(defined &bootstrap) {
     require DynaLoader;
     @Apache::Constants::ISA = qw(DynaLoader);
 }
 
-#XXX: should just generate all this from the documention =item's
+if(exists $ENV{MOD_PERL}) {
+    bootstrap Apache::Constants $Apache::Constants::VERSION;
+}
 
-my(@common)     = qw(OK DECLINED DONE NOT_FOUND FORBIDDEN
-		     AUTH_REQUIRED SERVER_ERROR);
-my(@methods)    = qw(M_CONNECT M_DELETE M_GET M_INVALID M_OPTIONS
-		     M_POST M_PUT M_TRACE METHODS);
-my(@options)    = qw(OPT_NONE OPT_INDEXES OPT_INCLUDES 
-		     OPT_SYM_LINKS OPT_EXECCGI OPT_UNSET OPT_INCNOEXEC
-		     OPT_SYM_OWNER OPT_MULTI OPT_ALL);
-my(@server)     = qw(MODULE_MAGIC_NUMBER
-		     SERVER_VERSION SERVER_SUBVERSION);
-my(@response)   = qw(DOCUMENT_FOLLOWS MOVED REDIRECT
-		     USE_LOCAL_COPY
-		     BAD_REQUEST
-		     BAD_GATEWAY 
-		     RESPONSE_CODES
-		     NOT_IMPLEMENTED
-		     NOT_AUTHORITATIVE
-		     CONTINUE);
-my(@satisfy)    = qw(SATISFY_ALL SATISFY_ANY SATISFY_NOSPEC);
-my(@remotehost) = qw(REMOTE_HOST REMOTE_NAME
-		     REMOTE_NOLOOKUP REMOTE_DOUBLE_REV);
-my(@http)       = qw(HTTP_METHOD_NOT_ALLOWED 
-		     HTTP_NOT_ACCEPTABLE 
-		     HTTP_LENGTH_REQUIRED
-		     HTTP_PRECONDITION_FAILED
-		     HTTP_SERVICE_UNAVAILABLE
-		     HTTP_VARIANT_ALSO_VARIES
-		     HTTP_NO_CONTENT
-		     HTTP_METHOD_NOT_ALLOWED 
-		     HTTP_NOT_ACCEPTABLE 
-		     HTTP_LENGTH_REQUIRED
-		     HTTP_PRECONDITION_FAILED
-		     HTTP_SERVICE_UNAVAILABLE
-		     HTTP_VARIANT_ALSO_VARIES);
-
-my $rc = [@common, @response];
-
-%Apache::Constants::EXPORT_TAGS = (
-    common     => \@common,
-    response   => $rc,
-    http       => \@http,
-    options    => \@options,
-    methods    => \@methods,
-    remotehost => \@remotehost,
-    satisfy    => \@satisfy,
-    server     => \@server,				   
-    #depreciated
-    response_codes => $rc,
-);
-
-@Apache::Constants::EXPORT_OK = (
-    @response,
-    @http,
-    @options,
-    @methods,
-    @remotehost,
-    @satisfy,
-    @server,
-); 
-   
-#*Apache::Constants::EXPORT = \@common;
-@Apache::Constants::EXPORT = (@common, qw(OPT_EXECCGI));
-
-eval { bootstrap Apache::Constants $Apache::Constants::VERSION; };
-if($@) {
-    die "$@\n" if exists $ENV{MOD_PERL};
-    warn "warning: can't `bootstrap Apache::Constants' outside of httpd\n";
+unless(defined &import) {
+    require Exporter;
+    require Apache::Constants::Exports;
+    *import = \&Exporter::import;
 }
 
 sub AUTOLOAD {
@@ -88,6 +23,21 @@ sub AUTOLOAD {
     goto &$Apache::Constants::AUTOLOAD;
 }
 
+my %ConstNameCache = ();
+
+sub name {
+    my($self, $const) = @_;
+    require Apache::Constants::Exports;
+    return $ConstNameCache{$const} if $ConstNameCache{$const};
+    
+    for (@Apache::Constants::EXPORT, 
+	 @Apache::Constants::EXPORT_OK) {
+	if ((\&{$_})->() eq $const) {
+	    return ($ConstNameCache{$const} = $_);
+	}
+    }
+}
+
 1;
 
 __END__
@@ -217,10 +167,10 @@ These are constants related to server version:
 
  MODULE_MAGIC_NUMBER
  SERVER_VERSION
- SERVER_SUBVERSION
+ SERVER_BUILT
 
 =back
 
 =head1 AUTHORS
 
-Gisle Aas <aas@sn.no>, Doug MacEachern <dougm@osf.org> and h2xs
+Doug MacEachern, Gisle Aas and h2xs
