@@ -2,12 +2,8 @@ package Apache::Registry;
 use Apache ();
 #use strict; #eval'd scripts will inherit hints
 use Apache::Constants qw(:common &OPT_EXECCGI &REDIRECT);
-use FileHandle ();
-use File::Basename ();
-use Cwd ();
 
-#$Id: Registry.pm,v 1.45 1998/02/10 08:58:18 dougm Exp $
-$Apache::Registry::VERSION = (qw$Revision: 1.45 $)[1];
+$Apache::Registry::VERSION = '2.00';
 
 $Apache::Registry::Debug ||= 0;
 # 1 => log recompile in errorlog
@@ -17,6 +13,13 @@ Apache->module('Apache::Debug') if $Apache::Registry::Debug;
 
 my $Is_Win32 = $^O eq "MSWin32";
 
+unless (defined $Apache::Registry::NameWithVirtualHost) {
+    $Apache::Registry::NameWithVirtualHost = 1;
+}
+unless (defined $Apache::Registry::MarkLine) {
+    $Apache::Registry::MarkLine = 1;
+}
+
 sub handler {
     my $r = shift;
     if(ref $r) {
@@ -51,17 +54,18 @@ sub handler {
 
 	my $mtime = -M _;
 
+	my $uri = $r->uri;
+	$uri = "/__INDEX__" if $uri eq "/";
 	# turn into a package name
 	$r->log_error(sprintf "Apache::Registry::handler examining %s",
-		      $r->uri) if $Debug && $Debug & 4;
+		      $uri) if $Debug && $Debug & 4;
 	my $script_name = $r->path_info ?
-	    substr($r->uri, 0, length($r->uri)-length($r->path_info)) :
-		$r->uri;
+	    substr($uri, 0, length($uri)-length($r->path_info)) :
+		$uri;
 
 	if($Apache::Registry::NameWithVirtualHost) {
-	    my $srv = $r->server;
-	    $script_name = join "", $srv->server_hostname, $script_name
-		if $srv->is_virtual;
+	    my $name = $r->get_server_name;
+	    $script_name = join "", $name, $script_name if $name;
 	}
 
 	# Escape everything into valid perl identifiers
@@ -69,7 +73,7 @@ sub handler {
 
 	# second pass cares for slashes and words starting with a digit
 	$script_name =~ s{
-			  (/)        # directory
+			  (/+)       # directory
 			  (\d?)      # package's first character
 			 }[
 			   "::" . ($2 ? sprintf("_%2x",unpack("C",$2)) : "")
@@ -80,9 +84,7 @@ sub handler {
 	$r->log_error("Apache::Registry::handler package $package")
 	   if $Debug && $Debug & 4;
 
-
-	my $cwd = Cwd::fastcwd();
-	chdir File::Basename::dirname($r->filename);
+	$r->chdir_file;
 
 	if (
 	    exists $Apache::Registry->{$package}{'mtime'}
@@ -93,13 +95,8 @@ sub handler {
  	} else {
 	    $r->log_error("Apache::Registry::handler reading $filename")
 		if $Debug && $Debug & 4;
-	    my($sub);
-	    {
-		my $fh = FileHandle->new($filename);
-		local $/;
-		$sub = <$fh>;
-		$sub = parse_cmdline($sub);
-	    }
+	    my $sub = $r->slurp_filename;
+	    $sub = parse_cmdline($$sub);
 
 	    # compile this subroutine into the uniq package name
             $r->log_error("Apache::Registry::handler eval-ing") if $Debug && $Debug & 4;
@@ -109,19 +106,24 @@ sub handler {
 	    }
 	    $r->clear_rgy_endav($script_name);
 
+	    my $line = $Apache::Registry::MarkLine ?
+		"\n#line 1 $filename\n" : "";
+ 
 	    my $eval = join(
 			    '',
 			    'package ',
 			    $package,
  			    ';use Apache qw(exit);',
  			    'sub handler {',
- 			    "\n#line 1 $filename\n",
+			    $line,
 			    $sub,
 			    "\n}", # last line comment without newline?
 			   );
 	    compile($eval);
+	    $r->stash_rgy_endav($script_name);
 	    if ($@) {
 		$r->log_error($@);
+		$@{$uri} = $@;
 		return SERVER_ERROR unless $Debug && $Debug & 2;
 		return Apache::Debug::dump($r, SERVER_ERROR);
 	    }
@@ -132,17 +134,14 @@ sub handler {
 
 	my $cv = \&{"$package\::handler"};
 	eval { &{$cv}($r, @_) } if $r->seqno;
-	{
-	    local $^W = 0; #shutup Cwd.pm
-	    chdir $cwd;
-	}
+	$r->chdir_file("$Apache::Server::CWD/");
 	$^W = $oldwarn;
 
 	my $errsv = "";
 	if($@) {
 	    $errsv = $@;
 	    $@ = ''; #XXX fix me, if we don't do this Apache::exit() breaks
-	    $@{"Apache::Registry"} = $errsv if $Apache::ERRSV_CAN_BE_HTTP;
+	    $@{$uri} = $errsv;
 	}
 
 	if($errsv) {
@@ -151,14 +150,12 @@ sub handler {
 	    return Apache::Debug::dump($r, SERVER_ERROR);
 	}
 
-=pod
-	#XXX
-	if(my $loc = $r->header_out("Location")) {
-	    if($r->status == 200 and substr($loc, 0, 1) ne "/") {
-		return REDIRECT;
-	    }
-	}
-=cut
+#	#XXX
+#	if(my $loc = $r->header_out("Location")) {
+#	    if($r->status == 200 and substr($loc, 0, 1) ne "/") {
+#		return REDIRECT;
+#	    }
+#	}
 	return $r->status;
     } else {
 	return NOT_FOUND unless $Debug && $Debug & 2;
@@ -175,7 +172,7 @@ sub compile {
 #XXX not good enough yet
 my(%switches) = (
    'T' => sub {
-       Apache::warn("Apache::Registry: -T switch ignored, ".
+       Apache::warn("Apache::Registry: T switch ignored, ".
 		    "enable with 'PerlTaintCheck On'\n")
 	   unless $Apache::__T; "";
    },
@@ -323,6 +320,5 @@ perl(1), mod_perl(3), Apache(3), Apache::Debug(3)
 
 =head1 AUTHORS
 
-Andreas Koenig <andreas.koenig@franz.ww.tu-berlin.de> and
-Doug MacEachern <dougm@osf.org>
+Andreas J. Koenig and Doug MacEachern
 
