@@ -3,7 +3,7 @@ use 5.003_97;
 use strict;
 
 BEGIN {
-    $mod_perl::VERSION = "1.0704";
+    $mod_perl::VERSION = "1.1501";
 }
 
 sub subversion {
@@ -19,6 +19,11 @@ sub hook {
     return Apache::perl_hook($try);
 }
 
+sub unimport {
+  my $class = shift;
+  %mod_perl::UNIMPORT = map { lc($_),1 } @_;
+}
+
 sub import {
     my $class = shift;
 
@@ -52,9 +57,6 @@ sub import {
     }
 }
 
-
-sub hooks { qw() }
-
 1;
 
 __END__
