@@ -1,8 +1,16 @@
+#PerlOpmask default
+
+<IfModule mod_dll.c>
+LoadModule perl_module modules/ApacheModulePerl.dll
+</IfModule>
+
 AddType text/x-server-parsed-html .shtml
 AddType text/perl-module .pm
 
 Action text/perl-module /perl/action.pl
 
+PerlPassEnv TEST_PERL_DIRECTIVES
+
 #these three are passed to perl_parse(), 
 #which happens before <Perl> sections are processed
 #optionally, they can be inside <Perl>, however for testing we want
@@ -21,15 +29,55 @@ PerlSetEnv KeyForPerlSetEnv OK
 
 <Perl>
 #!perl
+use Apache ();
+use Apache::Registry ();
+
+Apache::Server->register_cleanup(sub { 
+    warn "Apache::Server registered cleanup called for $$\n";
+});
+
+if($ENV{TEST_PERL_DIRECTIVES}) {
+    #t/TestDirectives/TestDirectives.pm
+    push @INC, map { "t/TestDirectives/blib/$_" } qw(arch lib);
+    require Apache::TestDirectives;
+    require Apache::ExtUtils;
+ 
+    my $proto_perl2c = Apache::ExtUtils->proto_perl2c;
+
+    $PerlConfig .= "YAC yet another\n";
+
+    $PerlConfig .= "<Location /perl>\n";
+    while(my($pp,$cp) = each %$proto_perl2c) {
+	my $arg = "A";
+	$pp =~ s/^\$\$//;
+	1 while $pp =~ s/(\$|\@)/$arg++ . " "/ge;
+	$PerlConfig .= "$cp $pp\n";
+    }
+
+    $PerlConfig .= <<EOF;
+TestCmd one two
+AnotherCmd
+CmdIterate A B C D E F
+</Location>
+<Container /for/whatever>
+
+it's  
+  miller
+time
+#make that a scotch
+</Container>
+
+<Location /perl/io>
+TestCmd PerlIO IsStdio
+</Location>
+EOF
+}
 
 $My::config_is_perl = 1;
 
-use Apache::Constants qw(MODULE_MAGIC_NUMBER);
-use IO::Handle ();
-use Cwd qw(fastcwd);
-my $dir = join "/", fastcwd, "t";
+my $dir = $Apache::Server::CWD;
+$dir .= "/t"; # if -d "t";
 my $Is_Win32 = ($^O eq "MSWin32");
-my $mmn = MODULE_MAGIC_NUMBER;
 
 sub prompt ($;$) {
     my($mess,$def) = @_;
@@ -55,13 +103,19 @@ if($User eq "root") {
 }
 print "Will run tests as User: '$User' Group: '$Group'\n";
 
-$Port = 8529;
+require 'net/config.pl';
+my $srv = $net::httpserver;
+($srv = $net::httpserver) =~ s/\D+$//;
+$Port = (split ":", $srv, 2)[1];
+$Port ||= 8529;
+print "Will bind to Port $Port\n";
+
 $DocumentRoot = "$dir/docs";
 $ServerName = "localhost";
 
 push @AddType, ["text/x-server-parsed-html" => ".shtml"];
  
-for (qw(/perl /cgi-bin)) {
+for (qw(/perl /cgi-bin /dirty-perl /perl_xs)) {
     push @Alias, [$_ => "$dir/net/perl/"];
 }
 
@@ -71,12 +125,32 @@ my @mod_perl = (
     Options     => "ExecCGI",
 );
 
-$Location{"/perl"} = { 
-    @mod_perl,
-    PerlSetEnv => [KeyForPerlSetEnv => "OK"],
-#    PerlSetVar => [KeyForPerlSetVar => "OK"],
+$Location{"/dirmagic"} = {
+    PerlHandler => "My::DirIndex",
 };
 
+$Location{"/dirty-perl"} = { 
+    SetHandler => "perl-script",
+    PerlHandler => "Apache::PerlRun",
+    Options => "+ExecCGI ",
+    PerlSendHeader => "On",
+};
+
+for (qw(perl perl_xs)) {
+    $Location{"/$_"} = { 
+	@mod_perl,
+	PerlSetEnv => [KeyForPerlSetEnv => "OK"],
+        #PerlSetVar => [KeyForPerlSetVar => "OK"],
+    };
+}
+
+for (qw(perl perl_xs)) {
+    $Location{"/$_/noenv"} = { 
+	@mod_perl,
+	PerlSetupEnv => "Off",
+    };
+}
+
 $Location{"/cgi-bin"} = {
     SetHandler => "cgi-script",
     Options    => "ExecCGI",
@@ -89,6 +163,11 @@ $VirtualHost{"localhost"} = {
 	    PerlSendHeader => "On",
 	    PerlSetupEnv   => "On",
 	},
+	"/perl_xs/io" => {
+	    @mod_perl,
+	    PerlSendHeader => "On",
+	    PerlSetupEnv   => "On",
+	},
     },
 };
 
@@ -98,10 +177,12 @@ $VirtualHost{"localhost"} = {
 #    PerlSetupEnv   => "On",
 #};
 
-$Location{"/perl/perl-status"} = {
-    SetHandler  => "perl-script",
-    PerlHandler => "Apache::Status",
-};
+for (qw(perl perl_xs)) {
+    $Location{"/$_/perl-status"} = {
+	SetHandler  => "perl-script",
+	PerlHandler => "Apache::Status",
+    };
+}
 
 for (qw(status info)) {
     $Location{"/server-$_"} = {
@@ -109,12 +190,12 @@ for (qw(status info)) {
     };
 }
 
-$ErrorLog = "/tmp/mod_perl_error_log";
-$PidFile  = "/tmp/mod_perl_httpd.pid";
+$ErrorLog = "logs/mod_perl_error_log";
+$PidFile  = "logs/mod_perl_httpd.pid";
 
-$AccessConfig = $TypesConfig = $TransferLog = $ScoreBoardFile = "/dev/null";
+$AccessConfig = $TypesConfig = $ScoreBoardFile = "$dir/docs/null.txt";
 
-$LockFile = "/tmp/mod_perl.lock";
+$LockFile = "logs/mod_perl.lock";
 
 #push @PerlChildInitHandler, "My::child_init";
 #push @PerlChildExitHandler, "My::child_exit";
@@ -139,3 +220,4 @@ $Location{"/chain"} = {
 };
 
 </Perl>
+
