@@ -4,9 +4,23 @@ BEGIN {
 
     use lib './t/docs';
     require "blib.pl" if -e "./t/docs/blib.pl";
-    $Apache::ServerStarting or die "Server is not starting !?\n";
+    $Apache::Server::Starting or warn "Server is not starting !?\n";
+    \$Apache::Server::Starting == \$Apache::ServerStarting or 
+	warn "GV alias broken\n";
+    \$Apache::Server::ReStarting == \$Apache::ServerReStarting or 
+	warn "GV alias broken\n";
 }
 
+use Apache ();
+use Apache::Registry ();
+
+#no mod_perl qw(Connection Server);
+
+eval {
+    require Apache::PerlRunXS;
+}; $@ = '' if $@;
+
+
 {
     last;
     Apache::warn("use Apache 'warn' is ok\n");
@@ -31,42 +45,29 @@ $Apache::ERRSV_CAN_BE_HTTP  = 1;
 #warn "ServerReStarting=$Apache::ServerReStarting\n";
 
 #use Apache::Debug level => 4;
-
 use mod_perl 1.03_01;
 
 if(defined &main::subversion) {
     die "mod_perl.pm is broken\n";
 }
 
-$ENV{KeyForPerlSetEnv} eq "OK" or warn "PerlSetEnv is broken\n";
-
-#test Apache::RegistryLoader
-{
-    use Cwd ();
-    use Apache::RegistryLoader ();
-    use DirHandle ();
-    use strict;
-    
-    local $^W = 0; #shutup line 164 Cwd.pm 
-
-    my $cwd = Cwd::fastcwd;
-    my $rl = Apache::RegistryLoader->new(trans => sub {
-	my $uri = shift; 
-	$cwd."/t/net${uri}";
-    });
-
-    my $d = DirHandle->new("t/net/perl");
-
-    for my $file ($d->read) {
-	next if $file eq "hooks.pl"; 
-	next unless $file =~ /\.pl$/;
-	my $status = $rl->handler("/perl/$file");
-	unless($status == 200) {
-	    die "pre-load of `/perl/$file' failed, status=$status\n";
-	}
-    }
+if($ENV{PERL_TEST_NEW_READ}) {
+    *Apache::READ = \&Apache::new_read;
 }
 
+$ENV{KeyForPerlSetEnv} eq "OK" or warn "PerlSetEnv is broken\n";
+
+%net::callback_hooks = ();
+require "net/config.pl";
+if($net::callback_hooks{PERL_SAFE_STARTUP}) {
+    eval "open \$0";
+    unless ($@ =~ /open trapped by operation mask/) {
+	die "opmask not set";
+    }
+}
+else {
+    require "./t/docs/rl.pl";
+}
 #for testing perl mod_include's
 
 $Access::Cnt = 0;
@@ -81,25 +82,39 @@ sub PerlTransHandler::handler {-1}
 #for testing PERL_HANDLER_METHODS
 #see httpd.conf and t/docs/LoadClass.pm
 
+use LoadClass ();
 sub MyClass::method ($$) {
     my($class, $r) = @_;  
-    warn "$class->method called\n";
+    #warn "$class->method called\n";
+    0;
 }
 
 sub BaseClass::handler ($$) {
     my($class, $r) = @_;  
-    warn "$class->handler called\n";
+    #warn "$class->handler called\n";
+    0;
 }
 
+{
+    package BaseClass;
+    #so 5.005-tobe doesn't complain:
+    #No such package "BaseClass" in @ISA assignment at ...
+}
+
+
+$MyClass::Object = bless {}, "MyClass";
 @MyClass::ISA = qw(BaseClass);
 
 #testing child init/exit hooks
 
 sub My::child_init {
     my $r = shift;
-    my $s = $r->server;
-    my $sa = $s->server_admin;
-    $s->warn("[notice] child_init for process $$, report any problems to $sa\n");
+    eval {
+      my $s = $r->server;
+      my $sa = $s->server_admin;
+      $s->warn("[notice] child_init for process $$, report any problems to $sa\n");
+    }; $@='' if $@;
+    0;
 }
 
 sub My::child_exit {
@@ -134,6 +149,45 @@ sub Apache::AuthenTest::handler {
     return OK;                       
 }
 
+use Apache::Constants qw(DECLINED DIR_MAGIC_TYPE);
+
+sub My::DirIndex::handler {
+    my $r = shift;
+    return DECLINED unless $r->content_type and 
+	$r->content_type eq DIR_MAGIC_TYPE;
+    require DirHandle;
+    my $dh = DirHandle->new($r->filename) or die $!;
+    my @entries = $dh->read;
+    my $x = @entries;
+    $r->send_http_header('text/plain');
+    print "1..$x\n";
+    my $i = 1;
+    for my $e (@entries) {
+	print "ok $i ($e)\n";
+	++$i;
+    }
+    1;
+}
+
+sub My::ProxyTest::handler {
+    my $r = shift;
+    unless ($r->proxyreq and $r->uri =~ /proxytest/) {
+	#warn sprintf "ProxyTest: proxyreq=%d, uri=%s\n",
+	$r->proxyreq, $r->uri;
+    }
+    return -1 unless $r->proxyreq;
+    return -1 unless $r->uri =~ /proxytest/;
+    $r->handler("perl-script");
+    $r->push_handlers(PerlHandler => sub {
+	my $r = shift;
+	$r->send_http_header("text/plain");
+	$r->print("1..1\n");
+	$r->print("ok 1\n");
+	$r->print("URI=`", $r->uri, "'\n");
+    });
+    return 0;
+}
+
 if(Apache->can_stack_handlers) {
     Apache->push_handlers(PerlChildExitHandler => sub {
 	warn "[notice] push'd PerlChildExitHandler called, pid=$$\n";
@@ -155,5 +209,4 @@ sub DESTROY {
 #prior to 1.3b1 (and the child_exit hook), this object's DESTROY method would not be invoked
 $global_object = Destruction->new;
 
-#0; #make sure we're not required to return a true value
 1;
